/*
 *   IgniteServ, include/timer.c
 *   Copyright (C) 2005 by those in Authors file
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "defs.h"

#include <unistd.h>
#include <sys/types.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#ifndef HAVE_CYGWIN
#include <signal.h>
#else
#include <sys/signal.h>
#endif /* HAVE_CYGWIN */

#include "chanserv.h"
#include "config.h"
#include "data.h"
#include "flood.h"
#include "gline.h"
#include "hybdefs.h"
#include "init.h"
#include "log.h"
#include "memoserv.h"
#include "nickserv.h"
#include "operserv.h"
#include "settings.h"
#include "sock.h"
#include "statserv.h"
#include "timer.h"
#include "jupe.h"

/* Global timer that will reduce local usage of time() -kre */
time_t current_ts;

/*
 * DoTimer()
 *
 * Execute various commands every second - such as checking for expired
 * temp glines etc.
 */
void DoTimer(time_t unixtime)
{
  if ((HubSock == -2) && ((unixtime % ConnectFrequency) == 0))
    CycleServers();

  /* exec this every 5 seconds */
  if ((unixtime % 5) == 0)
    {
      /*
       * Check for expired ignores
       */
      ExpireIgnores(unixtime);
    } /* if ((unixtime % 5) == 0) */

  if ((unixtime % 30) == 0)
    {
#ifdef NICKSERVICES

      /*
       * Check if there are any nick stealers to kill
       */
      CollisionCheck(unixtime);

#endif /* NICKSERVICES */

    } /* if ((unixtime % 30) == 0) */

  if ((unixtime % 60) == 0)
    {
#ifdef ALLOW_GLINES

      /*
       * Check for expired temp. glines every 1 minute
       */
      ExpireGlines(unixtime);
#endif

#if defined CHANNELSERVICES
      ExpireBans(unixtime);
	    ExpireAkicks(unixtime);
#endif

#ifdef JUPEVOTES
      /* expire server jupe votes */
      ExpireVotes(unixtime);
#endif

#if defined AUTO_ROUTING && defined SPLIT_INFO
      /* Check for splitted servers and optionally reconnect them -kre */
      ReconnectCheck(unixtime);
#endif

    } /* if ((unixtime % 60) == 0) */

  if ((unixtime % 3600) == 0)
    {
      /*
       * Check for expired nicknames, channels, and memos every hour
       */

#ifdef NICKSERVICES

      ExpireNicknames(unixtime);

#ifdef CHANNELSERVICES

      ExpireChannels(unixtime);
#endif

#ifdef MEMOSERVICES

      ExpireMemos(unixtime);
#endif

#endif /* NICKSERVICES */

#ifdef STATSERVICES
      /*
       * Check for any HostHash entries which have had 0 clients
       * for the past StatExpire days
       */
      ExpireStats(unixtime);
#endif

    } /* if ((unixtime % 3600) == 0) */

#if defined(NICKSERVICES) && defined(MEMOSERVICES)

  if (MemoPurgeFreq && ((unixtime % MemoPurgeFreq) == 0))
    {

      /*
       * Go through memo list and delete any marked for
       * deletion
       */
      PurgeCheck();
    }

#endif /* defined(NICKSERVICES) && defined(MEMOSERVICES) */

  /*
   * we have to add gmt_offset here or it won't be true at
   * midnight unless we are in the GMT timezone - for instance,
   * if we are in the EST timezone, we would be 5 hours earlier
   * than GMT, and thus ((unixtime % 86400) == 0) would only
   * be true at 19:00, so if we add gmt_offset (which would
   * be negative) to unixtime (at midnight), the value would
   * be the TS of 19:00, which would return true
   */
  if (((unixtime + gmt_offset) % 86400) == 0)
    {
#ifdef STATSERVICES
      /*
       * every day at midnight reset the daily max
       * user/oper/server/channel count
       */
      Network->MaxUsersT = Network->TotalUsers;
      Network->MaxOperatorsT = Network->TotalOperators;
      Network->MaxServersT = Network->TotalServers;
      Network->MaxChannelsT = Network->TotalChannels;

      Network->MaxUsersT_ts = unixtime;
      Network->MaxOperatorsT_ts = unixtime;
      Network->MaxServersT_ts = unixtime;
      Network->MaxChannelsT_ts = unixtime;

      Network->OperKillsT = 0;
      Network->ServKillsT = 0;

      putlog(LOG2, "%s: Reset daily counts",
             n_StatServ);
#endif /* STATSERVICES */

      /*
       * every day at midnight backup the log file
       * to the format: logfile.YYYYMMDD and start
       * a new one.
       */
      CheckLogs(unixtime);
    } /* if (((unixtime + gmt_offset) % 86400) == 0) */

  if (BackupFreq && (((unixtime + gmt_offset) % BackupFreq) == 0))
    {
      /*
       * Backup database files
       */
      BackupDatabases(unixtime);
    }

  if ((unixtime % DatabaseSync) == 0)
    {
      WriteDatabases();
    } /* if ((unixtime % DatabaseSync) == 0) */

#ifdef STATSERVICES

  /*
   * Every minute ping all servers to update lag information
   */
  if (LagDetect && ((unixtime % 60) == 0))
    {
      DoPings();
    }

#endif /* STATSERVICES */

  if (!SafeConnect && currenthub)
    {
      /*
       * currenthub->connect_ts + ConnectBurst is the TS
       * of when we can assume we are safely connected to
       * the network - that is, there is no more initial
       * connect burst information coming. So, we can now
       * broadcast clone/client connections to +y users
       * without fearing they will get flooded with hundreds
       * of them from an initial burst.
       * Make sure currenthub->realname is not null, so we
       * can tell if we are actually connected to a hub :-)
       */
      if (currenthub->realname &&
          (unixtime >= (currenthub->connect_ts + ConnectBurst)))
        SafeConnect = 1;
    } /* if (!SafeConnect && currenthub) */
} /* DoTimer() */
